/*
Module : FTPTransferDlg.cpp
Purpose: Defines the implementation for an MFC dialog which performs FTP uploads 
         and downloads similar to the Internet Explorer download dialog
Created: PJN / 03-01-2000
History: PJN / 18-10-2000 1. Added a m_bPromptOverwrite variable 
                          2. Fixed a bug in the way the progress control was being updated
         PJN / 11-06-2001 1. Updated copyright details.
                          2. Now provides Bandwidth throttling support. Thanks to Karim Mribti for
                          this good addition.
                          3. Minor code tidy up thanks to BoundsChecker
                          4. m_bPromptForOverwrite now works correctly for downloads as well as uploads
                          5. Fixed a problem with the progress control not updating correctly
         PJN / 10-10-2001 1. Fixed a problem where old style MFC exception handling was being used
                          instead of C++ standard.
         PJN / 17-10-2001 1. Now includes support for PASV
                          2. Fixed a Unicode issue in CFTPTransferDlg::OnStatusCallBack
                          3. Support for preconfigured or direct settings.
                          4. Support for proxies
         PJN / 02-02-2002 1. Now includes support for FTP resumes via the FTP command "REST". Please
                          note that IE 5 is required to support this. This support is experimental,
                          so your feedback is welcome.
                          2. Updated documentation to include explicit requirements for the class.
         PJN / 20-04-2002 1. Modified the code to correctly support FTP resumes. Feedback from 
                          jony for reporting this issue.
                          2. Fixed a bug in the call to AfxBeginThread
         PJN / 05-06-2002 1. Renamed some string resources used by the class to maintain naming 
                          consistency.
         PJN / 30-10-2002 1. Fixed a problem with seeking to the correct position when resuming a transfer. 
                          Thanks to Liping Dai for this bug report.
         PJN / 13-06-2003 1. Now supports connection timeouts since wininet doesn't <ggg>.
                          2. Other various tidy ups to the code, such as review of all the TRACE statements
                          and calls to GetLastError
         PJN / 20-11-2006 1. Updated the code to clean compile on VC 2005
                          2. Updated copyright details
                          3. Updated the code to compile correctly for Win64
                          4. Integrated the _FTP_DOWNLOAD_DATA class into the CFTPTransferDlg class
                          5. Optimized member variable construction in CFTPTransferDlg constructor
                          6. Code now uses newer C++ style casts in preference to C style casts
                          7. Refactored the code to produce a new CFTPTransferer class which provides
                          a synchronous interface to Wininet FTP which CFTPTransferDlg now inherits 
                          from. This is in line with the most recent updates in the author's CHttpDownloadDlg
                          class
                          8. Class now handles extra large files (> 4GB)
                          9. Code now correctly handles ERROR_INTERNET_EXTENDED_ERROR errors
         PJN / 10-06-2007 1. Minor bug fix to CFTPTransferDlg::SetTimeLeft to correctly handle 64 bit integer
                          parameter.
         PJN / 18-05-2008 1. Updated copyright details
                          2. Updated the logic in CFTPTransferDlg::OnWininetStatusCallBack to correctly handle 
                          ASCII or Unicode strings. Thanks to Hans Dietrich for prompting this update.
         PJN / 24-01-2017 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2010 to 2015
                          3. Replaced BOOL with bool throughout the codebase
                          4. Updated the code to clean compile using /analyze
                          5. Replaced NULL with nullptr throughout the codebase. This means that the code now 
                          requires VC 2010 at a minimum to compile
                          6. Implemented support for send and receive timeouts
         PJN / 06-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
                          3. Changed enums to enum class throughout codebase
                          4. Reworked all the notifications from the worker thread in CFTPTransferDlg to the UI
                          to use PostMessage
         PJN / 29-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 27-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 17-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 2000 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "FTPTransferDlg.h"


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

constexpr UINT WM_FTPTRANSFER_THREAD_FINISHED = WM_APP + 1;
constexpr UINT WM_FTPTRANSFER_THREAD_UIUPDATE = WM_APP + 2;


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440 26477)
IMPLEMENT_DYNAMIC(CFTPTransferDlg, CDialog);

#pragma warning(suppress: 26455)
CFTPTransferDlg::CFTPTransferDlg(_In_opt_ CWnd* pParent) : CDialog{CFTPTransferDlg::IDD, pParent},
                                                           m_bSafeToClose{false},
                                                           m_nLastTotalBytes{0},
                                                           m_dwLastPercentage{0xFFFFFFFF},
                                                           m_nLastTicks{0},
                                                           m_nStartTicks{0},
                                                           m_pThread{nullptr},
                                                           m_UIUpdate{UIUpdate::None}
{
}

void CFTPTransferDlg::DoDataExchange(CDataExchange* pDX)
{
  __super::DoDataExchange(pDX);
  DDX_Control(pDX, IDC_STATUS, m_ctrlStatus);
  DDX_Control(pDX, IDC_TRANSFER_RATE, m_ctrlTransferRate);
  DDX_Control(pDX, IDC_TIMELEFT, m_ctrlTimeLeft);
  DDX_Control(pDX, IDC_PROGRESS1, m_ctrlProgress);
  DDX_Control(pDX, IDC_FILESTATUS, m_ctrlFileStatus);
  DDX_Control(pDX, IDC_ANIMATE1, m_ctrlAnimate);
}

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CFTPTransferDlg, CDialog) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_DESTROY()
  ON_WM_CLOSE()
  ON_MESSAGE(WM_FTPTRANSFER_THREAD_FINISHED, &CFTPTransferDlg::OnThreadFinished)
  ON_MESSAGE(WM_FTPTRANSFER_THREAD_UIUPDATE, &CFTPTransferDlg::OnUIUpdate)
END_MESSAGE_MAP()

LRESULT CFTPTransferDlg::OnThreadFinished(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  //It's now safe to close since the thread has signaled us
  m_bSafeToClose = true;

  //Stop the animation
  m_ctrlAnimate.Stop();

  //If an error occurred display the message box
  if (IsAbort())
    EndDialog(IDCANCEL);
  else if (m_sError.GetLength())
  {
    AfxMessageBox(m_sError);
    EndDialog(IDCANCEL);
  }
  else
    EndDialog(IDOK);

  return 0L;
}

BOOL CFTPTransferDlg::OnInitDialog()
{
  //Let the parent class do its thing
  __super::OnInitDialog();

  //Empty out the error string
  m_sError.Empty();

  //Reset the values we use for updating the progress
#pragma warning(suppress: 26472)
  m_dwLastPercentage = static_cast<DWORD>(-1);
  m_nLastTotalBytes = 0;

  //Change the caption of the dialog if necessary
  if (!m_bDownload)
  {
    CString sCaption;
    if (sCaption.LoadString(IDS_FTPTRANSFER_UPLOAD_CAPTION))
      SetWindowText(sCaption);
  }

  //Set the range of the progress control
  m_ctrlProgress.SetRange(0, 100);

  //Setup the animation control
  m_ctrlAnimate.Open(IDR_FTPTRANSFER_ANIMATION);

  //Set the file status text
  CString sFileStatus;
  ASSERT(m_sRemoteFile.length());
  ASSERT(m_sServer.length());
  if (m_bDownload)
    AfxFormatString2(sFileStatus, IDS_FTPTRANSFER_FILESTATUS1, m_sRemoteFile.c_str(), m_sServer.c_str());
  else
    AfxFormatString2(sFileStatus, IDS_FTPTRANSFER_FILESTATUS2, m_sRemoteFile.c_str(), m_sServer.c_str());
  m_ctrlFileStatus.SetWindowText(sFileStatus);

  //Start the animation to signify that the download is taking place
#pragma warning(suppress: 26472)
  m_ctrlAnimate.Play(0, static_cast<UINT>(-1), static_cast<UINT>(-1));

  //Spin off the background thread which will do the actual file transfer
  m_pThread = AfxBeginThread(_TransferThread, this, THREAD_PRIORITY_NORMAL, 0, CREATE_SUSPENDED);
  if (m_pThread == nullptr)
  {
    TRACE(_T("CFTPTransferDlg::OnInitDialog, Failed to create transfer thread, dialog is aborting\n"));
    EndDialog(IDCANCEL);
    return TRUE;
  }
  m_pThread->m_bAutoDelete = FALSE;
  m_pThread->ResumeThread();

  return TRUE;
}

UINT CFTPTransferDlg::_TransferThread(_In_ LPVOID pParam)
{
  //Convert from the SDK world to the C++ world
  auto pDlg{static_cast<CFTPTransferDlg*>(pParam)};
#pragma warning(suppress: 26496)
  AFXASSUME(pDlg != nullptr);
  pDlg->TransferThread();
  return 0;
}

void CFTPTransferDlg::SetPercentage(_In_ int nPercentage)
{
  //Change the progress control
  m_ctrlProgress.SetPos(nPercentage);

  //Change the caption text
  CString sPercentage;
  sPercentage.Format(_T("%d"), nPercentage);
  CString sCaption;
  AfxFormatString2(sCaption, IDS_FTPTRANSFER_PERCENTAGE, sPercentage, m_sRemoteFile.c_str());
  {
    ATL::CCritSecLock sl{m_csUIUpdateText.m_sec, true};
    m_UIUpdate = UIUpdate::Percentage;
    m_sUIUpdateText = sCaption;
  }
  PostMessage(WM_FTPTRANSFER_THREAD_UIUPDATE);
}

void CFTPTransferDlg::SetTimeLeft(_In_ DWORD dwSecondsLeft, _In_ ULONGLONG dwBytesRead, _In_ ULONGLONG dwFileSize)
{
  CString sCopied;
  if (dwBytesRead < 1024)
  {
    CString sBytes;
    sBytes.Format(_T("%I64u"), dwBytesRead);
    AfxFormatString1(sCopied, IDS_FTPTRANSFER_BYTES, sBytes);
  }
  else if (dwBytesRead < 1048576)
  {
    CString sKiloBytes;
    sKiloBytes.Format(_T("%0.1f"), dwBytesRead/1024.0);
    AfxFormatString1(sCopied, IDS_FTPTRANSFER_KILOBYTES, sKiloBytes);
  }
  else
  {
    CString sMegaBytes;
    sMegaBytes.Format(_T("%0.2f"), dwBytesRead/1048576.0);
    AfxFormatString1(sCopied, IDS_FTPTRANSFER_MEGABYTES, sMegaBytes);
  }

  CString sTotal;
  if (dwFileSize < 1024)
  {
    CString sBytes;
    sBytes.Format(_T("%I64u"), dwFileSize);
    AfxFormatString1(sTotal, IDS_FTPTRANSFER_BYTES, sBytes);
  }
  else if (dwFileSize < 1048576)
  {
    CString sKiloBytes;
    sKiloBytes.Format(_T("%0.1f"), dwFileSize/1024.0);
    AfxFormatString1(sTotal, IDS_FTPTRANSFER_KILOBYTES, sKiloBytes);
  }
  else
  {
    CString sMegaBytes;
    sMegaBytes.Format(_T("%0.2f"), dwFileSize/1048576.0);
    AfxFormatString1(sTotal, IDS_FTPTRANSFER_MEGABYTES, sMegaBytes);
  }

  CString sOf;
  AfxFormatString2(sOf, IDS_FTPTRANSFER_OF, sCopied, sTotal);

  CString sTime;
  if (dwSecondsLeft < 60)
  {
    CString sSeconds;
    sSeconds.Format(_T("%u"), dwSecondsLeft);
    AfxFormatString1(sTime, IDS_FTPTRANSFER_SECONDS, sSeconds);
  }
  else
  {
    const DWORD dwMinutes{dwSecondsLeft / 60};
    const DWORD dwSeconds{dwSecondsLeft % 60};
    CString sSeconds;
    sSeconds.Format(_T("%u"), dwSeconds);
    CString sMinutes;
    sMinutes.Format(_T("%u"), dwMinutes);
    if (dwSeconds == 0)
      AfxFormatString1(sTime, IDS_FTPTRANSFER_MINUTES, sMinutes);
    else
      AfxFormatString2(sTime, IDS_FTPTRANSFER_MINUTES_AND_SECONDS, sMinutes, sSeconds);
  }
  CString sTimeLeft;
  AfxFormatString2(sTimeLeft, IDS_FTPTRANSFER_TIMELEFT, sTime, sOf);
  {
    ATL::CCritSecLock sl{m_csUIUpdateText.m_sec, true};
    m_UIUpdate = UIUpdate::TimeLeft;
    m_sUIUpdateText = sTimeLeft;
  }
  PostMessage(WM_FTPTRANSFER_THREAD_UIUPDATE);
}

void CFTPTransferDlg::SetTransferRate(_In_ double KbPerSecond)
{
  CString sRate;
  if (KbPerSecond < 1)
  {
    CString sBytesPerSecond;
    sBytesPerSecond.Format(_T("%0.0f"), KbPerSecond*1024);
    AfxFormatString1(sRate, IDS_FTPTRANSFER_BYTESPERSECOND, sBytesPerSecond);
  }
  else if (KbPerSecond < 10)
  {
    CString sKiloBytesPerSecond;
    sKiloBytesPerSecond.Format(_T("%0.2f"), KbPerSecond);
    AfxFormatString1(sRate, IDS_FTPTRANSFER_KILOBYTESPERSECOND, sKiloBytesPerSecond);
  }
  else
  {
    CString sKiloBytesPerSecond;
    sKiloBytesPerSecond.Format(_T("%0.0f"), KbPerSecond);
    AfxFormatString1(sRate, IDS_FTPTRANSFER_KILOBYTESPERSECOND, sKiloBytesPerSecond);
  }
  {
    ATL::CCritSecLock sl{m_csUIUpdateText.m_sec, true};
    m_UIUpdate = UIUpdate::Rate;
    m_sUIUpdateText = sRate;
  }
  PostMessage(WM_FTPTRANSFER_THREAD_UIUPDATE);
}

void CFTPTransferDlg::TransferThread()
{
  //Call the function which does the heavy lifting
  Transfer();

  //We're finished
  PostMessage(WM_FTPTRANSFER_THREAD_FINISHED);
}

#pragma warning(suppress: 26434)
void CFTPTransferDlg::OnDestroy()
{
  //Wait for the worker thread to exit
  if (m_pThread != nullptr)
  {
    WaitForSingleObject(m_pThread->m_hThread, INFINITE);
    delete m_pThread;
    m_pThread = nullptr;
  }

  //Let the parent class do its thing
  __super::OnDestroy();
}

void CFTPTransferDlg::OnCancel()
{
  //Just set the abort flag and disable the cancel button
  SetAbort();
  GetDlgItem(IDCANCEL)->EnableWindow(FALSE);
  __super::SetStatus(IDS_FTPTRANSFER_ABORTING_TRANSFER);
}

#pragma warning(suppress: 26434)
void CFTPTransferDlg::OnClose()
{
  if (m_bSafeToClose)
    __super::OnClose();
  else
  {
    //Just set the abort flag and disable the cancel button
    SetAbort();
    GetDlgItem(IDCANCEL)->EnableWindow(FALSE);
    __super::SetStatus(IDS_FTPTRANSFER_ABORTING_TRANSFER);
  }
}

void CFTPTransferDlg::SetStatus(_In_z_ LPCTSTR pszStatus)
{
  //For debugging purposes also run CFTPTransferer::SetStatus
  __super::SetStatus(pszStatus);

  //Simply update the status controls text
  {
    ATL::CCritSecLock sl{m_csUIUpdateText.m_sec, true};
    m_UIUpdate = UIUpdate::Status;
    m_sUIUpdateText = pszStatus;
  }
  PostMessage(WM_FTPTRANSFER_THREAD_UIUPDATE);
}

void CFTPTransferDlg::HandleError(_In_ UINT nID, _In_ DWORD dwError)
{
  //For debugging purposes also run CFTPTransferer::HandleError
  __super::HandleError(nID, dwError);

  //Form the error string to report
  CString sError;
  if (dwError == ERROR_INTERNET_EXTENDED_ERROR)
  {
    DWORD dwInetError{0};
    DWORD dwSize{0};
    ::InternetGetLastResponseInfo(&dwInetError, nullptr, &dwSize);
    TCHAR* pszResponse{sError.GetBuffer(dwSize)};
    ::InternetGetLastResponseInfo(&dwInetError, pszResponse, &dwSize);
    sError.ReleaseBuffer();
  }
  else
    sError = GetErrorMessage(dwError).c_str();
  AfxFormatString1(m_sError, nID, sError);
}

void CFTPTransferDlg::OnProgress(_In_ ULONGLONG nTotalBytesRead, _In_ bool bGotFileSize, _In_ ULONGLONG nFileSize)
{
  //For debugging purposes also run CHttpDownloader::OnProgress
  __super::OnProgress(nTotalBytesRead, bGotFileSize, nFileSize);

  if (m_dwLastPercentage == 0xFFFFFFFF) //Is this the first time we have been called for this download
  {
    m_nStartTicks = GetTickCount64();
    m_nLastTicks = m_nStartTicks;
  }

  if (bGotFileSize)
  {
    //Update the percentage downloaded in the caption
#pragma warning(suppress: 26472)
    const auto dwPercentage{static_cast<DWORD>(nTotalBytesRead * 100 / nFileSize)};
    if (dwPercentage != m_dwLastPercentage)
    {
      //Update the progress control bar
      SetPercentage(dwPercentage);
      m_dwLastPercentage = dwPercentage;
    }
  }

  //Update the transfer rate amd estimated time left every second
  const ULONGLONG nNowTicks{GetTickCount64()};
  const ULONGLONG nTimeTaken{nNowTicks - m_nLastTicks};
  if (nTimeTaken > 1000) //By default, only update the transfer rate and time left every second
  {
    const double KbPerSecond{(static_cast<double>(nTotalBytesRead) - static_cast<double>(m_nLastTotalBytes)) / nTimeTaken};
    SetTransferRate(KbPerSecond);

    //Setup for the next time around the loop
    m_nLastTicks = nNowTicks;
    m_nLastTotalBytes = nTotalBytesRead;

    if (bGotFileSize)
    {
      //Update the estimated time left if we have downloaded something so far
      if (nTotalBytesRead)
      {
        const auto dwSecondsLeft{static_cast<int>((static_cast<double>(nNowTicks) - m_nStartTicks) / static_cast<double>(nTotalBytesRead) * (static_cast<double>(nFileSize) - static_cast<double>(nTotalBytesRead)) / 1000)};
        SetTimeLeft(dwSecondsLeft, nTotalBytesRead, nFileSize);
      }
    }
  }
}

void CFTPTransferDlg::OnWininetStatusCallBack(_In_ HINTERNET /*hInternet*/, _In_ DWORD dwInternetStatus, _In_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength)
{
  //Note the MSDN documentation has very poor information on what the lpStatusInformation parameter actually contains for various callbacks. Some users of 
  //the class has reported that it reports a LPTSTR for string type information, but the author has seen it return a LPSTR even for Unicode builds. As a 
  //defense in depth approach against this uncertainty, lets determine at runtime what type of string we have. For more info on this, check out 
  //http://groups.google.com/group/microsoft.public.inetsdk.programming.wininet/browse_thread/thread/e75e20aa425cfb16/a569b0026815bc98

  switch (dwInternetStatus)
  {
    case INTERNET_STATUS_RESOLVING_NAME:
    {
      //Determine if it is ascii or unicode text
      const BOOL bUnicodeText{IsTextUnicode(lpvStatusInformation, dwStatusInformationLength, nullptr)};

      //Just delegate back up to the base class to update the status text
      if (bUnicodeText)
        __super::SetStatus(IDS_FTPTRANSFER_RESOLVING_NAME, CString{static_cast<LPWSTR>(lpvStatusInformation)});
      else
        __super::SetStatus(IDS_FTPTRANSFER_RESOLVING_NAME, CString{static_cast<LPSTR>(lpvStatusInformation)});
      break;
    }
    case INTERNET_STATUS_NAME_RESOLVED:
    {
      //Determine if it is ascii or unicode text
      const BOOL bUnicodeText{IsTextUnicode(lpvStatusInformation, dwStatusInformationLength, nullptr)};

      if (bUnicodeText)
        __super::SetStatus(IDS_FTPTRANSFER_RESOLVED_NAME, CString{static_cast<LPWSTR>(lpvStatusInformation)});
      else
        __super::SetStatus(IDS_FTPTRANSFER_RESOLVED_NAME, CString{static_cast<LPSTR>(lpvStatusInformation)});
      break;
    }
    case INTERNET_STATUS_CONNECTING_TO_SERVER:
    {
      //Determine if it is ascii or unicode text
      const BOOL bUnicodeText{IsTextUnicode(lpvStatusInformation, dwStatusInformationLength, nullptr)};

      //Just delegate back up to the base class to update the status text
      if (bUnicodeText)
        __super::SetStatus(IDS_FTPTRANSFER_CONNECTING, CString{static_cast<LPWSTR>(lpvStatusInformation)});
      else
        __super::SetStatus(IDS_FTPTRANSFER_CONNECTING, CString{static_cast<LPSTR>(lpvStatusInformation)});
      break;
    }
    case INTERNET_STATUS_CONNECTED_TO_SERVER:
    {
      //Determine if it is ascii or unicode text
      const BOOL bUnicodeText{IsTextUnicode(lpvStatusInformation, dwStatusInformationLength, nullptr)};

      //Just delegate back up to the base class to update the status text
      if (bUnicodeText)
        __super::SetStatus(IDS_FTPTRANSFER_CONNECTED, CString{static_cast<LPWSTR>(lpvStatusInformation)});
      else
        __super::SetStatus(IDS_FTPTRANSFER_CONNECTED, CString{static_cast<LPSTR>(lpvStatusInformation)});
      break;
    }
    default:
    {
      break;
    }
  }
}

LRESULT CFTPTransferDlg::OnUIUpdate(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  ATL::CCritSecLock sl{m_csUIUpdateText.m_sec, true};
  switch (m_UIUpdate)
  {
    case UIUpdate::Percentage:
    {
      SetWindowText(m_sUIUpdateText);
      break;
    }
    case UIUpdate::TimeLeft:
    {
      m_ctrlTimeLeft.SetWindowText(m_sUIUpdateText);
      break;
    }
    case UIUpdate::Rate:
    {
      m_ctrlTransferRate.SetWindowText(m_sUIUpdateText);
      break;
    }
    case UIUpdate::Status:
    {
      m_ctrlStatus.SetWindowText(m_sUIUpdateText);
      break;
    }
    default:
    {
      break;
    }
  }

  return 0L;
}
