/*
Module : FTPTransferer.cpp
Purpose: Defines the implementation for an MFC class which provides a wrapping for downloading HTTP requests in a 
         synchronous manner via Wininet
Created: PJN / 15-07-2006
History: PJN / 13-04-2007 1. Updated copyright details.
                          2. Following feedback from Philippe MARINUTTI, it looks like you just cannot do
                          FTP resumes using Wininet. For more information on the issues involved, please see
                          http://groups.google.com/group/microsoft.public.inetsdk.programming.wininet/browse_thread/thread/b43094df4de38a35/376b4c14acdb9be6?lnk=st&q=FtpCommand+REST&rnum=9&hl=en#376b4c14acdb9be6,
                          http://groups.google.com/group/microsoft.public.inetsdk.programming.wininet/browse_thread/thread/229a31ebda334098/4facc1ebcc9a7c2e?lnk=st&q=FtpCommand+REST&rnum=6&hl=en#4facc1ebcc9a7c2e,
                          http://groups.google.com/group/microsoft.public.inetsdk.programming.wininet/browse_thread/thread/5f1a1d5c92e6548d/1dfcc8fb46c6c721?lnk=st&q=FtpCommand+REST&rnum=10&hl=en#1dfcc8fb46c6c721 & 
                          http://groups.google.com/group/microsoft.public.inetsdk.programming.wininet/browse_thread/thread/31a840505f3b45bc/8c1732d682ba4f7b?lnk=st&q=FtpCommand+REST&rnum=1&hl=en#8c1732d682ba4f7b.
                          As such I have removed support for resumes (CFTPTransferer::m_dwStartPos) from the code.
                          If you would like to support FTP resumes, then you will need to look to commercial alternatives such as 
                          SmartFTP (http://www.smartftp.com/ftplib/), Xceed FTP Library (https://xceed.com/xceed-ftp-for-activex/),
                          or FTP Client Engine for C/C++ (http://www.marshallsoft.com/fce4c.htm).
         PJN / 30-12-2007 1. Updated the code to remove VC 6 style appwizard comments.
                          2. Updated the sample apps to clean compile on VC 2008.
                          3.  CFTPTransferer::GetErrorMessage now uses the FORMAT_MESSAGE_IGNORE_INSERTS flag. For more information please see Raymond Chen's blog at 
                          http://blogs.msdn.com/oldnewthing/archive/2007/11/28/6564257.aspx. Thanks to Alexey Kuznetsov for reporting this issue.
                          4. CFTPTransferer::GetErrorMessage now uses Checked::tcsncpy_s if compiled using VC 2005 or later.
         PJN / 18-05-2008 1. Updated copyright details
                          2. Addition of AttachSession and DetachSession methods which allow the lifetime of the session
                          to be controlled independently of the lifetime of the CHttpDownloader instance. Thanks to 
                          Hans Dietrich for prompting this update
                          3. Fixed a spelling mistake in the IDS_FTPTRANSFER_RETRIEVING_FILE string resource
         PJN / 24-01-2017 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2010 to 2015
                          3. Replaced BOOL with bool throughout the codebase
                          4. Updated the code to clean compile using /analyze
                          5. Replaced NULL with nullptr throughout the codebase. This means that the code now requires VC 2010 at a minimum to compile
                          6. Implemented support for send and receive timeouts
         PJN / 05-10-2017 1. Replaced CString::operator LPC*STR() calls throughout the codebase with CString::GetString calls
         PJN / 23-12-2017 1. Fix compile problems in CFTPTransferer::CreateSession in VC2017
                          2. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is defined.
         PJN / XX-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
                          3. Changed enums to enum class throughout codebase
                          4. Reworked the CFTPTransferer class to be independent of MFC
         PJN / 06-05-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 29-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 27-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 04-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 12-12-2021 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 16-04-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 2006 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "FTPTransferer.h"
#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (usually stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

//Link to the wininet lib file automatically
#pragma comment(lib, "Wininet.lib")


//////////////////// Implementation ///////////////////////////////////////////

CFTPTransferer::CFTPTransferer() noexcept : m_bDownload{true},
                                            m_nPort{INTERNET_DEFAULT_FTP_PORT},
                                            m_bBinary{true},
                                            m_bPasv{false},
                                            m_dbLimit{0},
                                            m_dwConnectionTimeout{0},
                                            m_ConnectionType{ConnectionType::UsePreConfig},
                                            m_dwReadBufferSize{1024},
                                            m_dwSendTimeout{0},
                                            m_dwReceiveTimeout{0},
                                            m_hInternetSession{nullptr},
                                            m_hFTPConnection{nullptr},
                                            m_hFTPFile{nullptr},
                                            m_bAbort{false},
                                            m_dwTempError{0}
{
}

CFTPTransferer::~CFTPTransferer()
{
#pragma warning(suppress: 26447)
  ReleaseHandles();
}

void CFTPTransferer::AttachSession(_In_ HINTERNET hInternetSession) noexcept
{
  if (m_hInternetSession != nullptr)
    ::InternetCloseHandle(m_hInternetSession);
  m_hInternetSession = hInternetSession;
}

HINTERNET CFTPTransferer::DetachSession() noexcept
{
  HINTERNET hTempSession{m_hInternetSession};
  m_hInternetSession = nullptr;
  return hTempSession;
}

void CFTPTransferer::ReleaseHandles()
{
  //Serialize access to the handles
  ATL::CCritSecLock sl{m_csHandles.m_sec, true};

  //Free up the internet handles we may be using
  if (m_hFTPFile != nullptr)
  {
    ::InternetCloseHandle(m_hFTPFile);
    m_hFTPFile = nullptr;
  }
  if (m_hFTPConnection != nullptr)
  {
    ::InternetCloseHandle(m_hFTPConnection);
    m_hFTPConnection = nullptr;
  }
  if (m_hInternetSession != nullptr)
  {
    ::InternetCloseHandle(m_hInternetSession);
    m_hInternetSession = nullptr;
  }
}

#pragma warning(suppress: 26440)
bool CFTPTransferer::DeleteDownloadedFile()
{
  const bool bSuccess{::DeleteFile(m_sLocalFile.c_str()) != FALSE};
  if (!bSuccess)
    ATLTRACE(_T("CFTPTransferer::DeleteDownloadedFile, Failed to delete failed downloaded file, Error:%u\n"), ::GetLastError());
  return bSuccess;
}

bool CFTPTransferer::Transfer()
{
  //Remember if we have the file to download into before we do the download
  const BOOL bDownloadFileAlreadyExists{GetFileAttributes(m_sLocalFile.c_str()) != INVALID_FILE_ATTRIBUTES};

  //Call the internal version which does the heavy lifting
  const bool bSuccess{_Transfer()};

  //Free up the internet handles now that we are finished with them
  ReleaseHandles();

  //Close the local file
  if (m_LocalFile.operator HANDLE() != nullptr)
    m_LocalFile.Close();

  //Remove any downloaded file which did not come down successfully (when it did not exist before we did the download) 
  if (m_bDownload && !bSuccess && !bDownloadFileAlreadyExists && (GetFileAttributes(m_sLocalFile.c_str()) != INVALID_FILE_ATTRIBUTES))
    DeleteDownloadedFile();

  return bSuccess;
}

bool CFTPTransferer::CreateSession()
{
  //Serialize access to the handles
  ATL::CCritSecLock sl{m_csHandles.m_sec, true};

  //Create the Internet session handle
  if (m_hInternetSession == nullptr)
  {
    switch (m_ConnectionType)
    {
      case ConnectionType::UsePreConfig:
      {
        m_hInternetSession = ::InternetOpen(m_sUserAgent.length() ? m_sUserAgent.c_str() : nullptr, INTERNET_OPEN_TYPE_PRECONFIG, nullptr, nullptr, 0);
        break;
      }
      case ConnectionType::DirectToInternet:
      {
        m_hInternetSession = ::InternetOpen(m_sUserAgent.length() ? m_sUserAgent.c_str() : nullptr, INTERNET_OPEN_TYPE_DIRECT, nullptr, nullptr, 0);
        break;
      }
      case ConnectionType::UseProxy:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(m_sProxyServer.length()); //A proxy server must be specified
        m_hInternetSession = ::InternetOpen(m_sUserAgent.length() ? m_sUserAgent.c_str() : nullptr, INTERNET_OPEN_TYPE_PROXY, m_sProxyServer.c_str(), nullptr, 0);
        break;
      }
      default:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(false);
        break;
      }
    }
  }

  if (m_hInternetSession == nullptr)
  {
    const DWORD dwError{GetLastError()};
    sl.Unlock();
    ATLTRACE(_T("CFTPTransferer::CreateSession, Failed in call to InternetOpen, Error:%u\n"), dwError);
    HandleError(IDS_FTPTRANSFER_GENERIC_ERROR, dwError);
    return false;
  }

  //Setup the status callback function
  if (::InternetSetStatusCallback(m_hInternetSession, _OnWininetStatusCallBack) == INTERNET_INVALID_STATUS_CALLBACK)
  {
    const DWORD dwError{GetLastError()};
    sl.Unlock();
    ATLTRACE(_T("CFTPTransferer::Download, Failed in call to InternetSetStatusCallback, Error:%u\n"), dwError);
    HandleError(IDS_FTPTRANSFER_GENERIC_ERROR, dwError);
    return false;
  }

  return true;
}

bool CFTPTransferer::CreateConnection()
{
  //Serialize access to the handles
  ATL::CCritSecLock sl{m_csHandles.m_sec, true};

  //Make the connection to the FTP server
#pragma warning(suppress: 26477)
  ATLASSERT(m_hFTPConnection == nullptr);

  //Should we try to call InternetConnect synchronously or use a worker thread to make the connection on our behalf
  if (m_dwConnectionTimeout == 0)
    MakeConnection();
  else
  {
    //Spin of the thread which does the connection for us. Note we use a worker thread so that we avoid
    //the blocking behaviour which FTP Wininet suffers from and which cannot be easily fixed.
#pragma warning(suppress: 26490)
    auto hWorkerThread{reinterpret_cast<HANDLE>(_beginthreadex(nullptr, 0, _MakeConnectionThread, this, CREATE_SUSPENDED, nullptr))};
    if (hWorkerThread == nullptr)
    {
      const DWORD dwError{GetLastError()};
      sl.Unlock();
      ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed to create the connection thread, Error:%u\n"), dwError);
      HandleError(IDS_FTPTRANSFER_FAIL_CONNECT_SERVER, dwError);
      return false;
    }

    //Everything ok, Resume the thread
#pragma warning(suppress: 26486)
    if (ResumeThread(hWorkerThread) == ULONG_MAX)
    {
      const DWORD dwError{GetLastError()};
      sl.Unlock();
      ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed in resume the connection thread, Error:%u\n"), dwError);
      HandleError(IDS_FTPTRANSFER_FAIL_CONNECT_SERVER, dwError);
      return false;
    }

    //Wait for the thread to return or a timeout occurs
#pragma warning(suppress: 26486)
    if (WaitForSingleObject(hWorkerThread, m_dwConnectionTimeout) == WAIT_TIMEOUT)
    {
      //Close the wininet session handle (which will cause the worker thread to return from its blocking call which 
      //most likely will be the call to InternetConnect)
      ::InternetCloseHandle(m_hInternetSession);
      m_hInternetSession = nullptr;
      sl.Unlock();

      //Wait until the worker thread exits, should be quick since we have just closed the session handle above!
#pragma warning(suppress: 26486)
      WaitForSingleObject(hWorkerThread, INFINITE);
#pragma warning(suppress: 26486)
      CloseHandle(hWorkerThread);

      ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed to connect to the FTP server in a timely manner\n"));
      HandleError(IDS_FTPTRANSFER_FAIL_CONNECT_SERVER, ERROR_TIMEOUT);

      return false;
    }

#pragma warning(suppress: 26486)
    CloseHandle(hWorkerThread);
  }

  //Check to see if the connection was successful
  if (m_hFTPConnection == nullptr)
  {
    ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed in call to InternetConnect, Error:%u\n"), m_dwTempError);
    sl.Unlock();
    HandleError(IDS_FTPTRANSFER_FAIL_CONNECT_SERVER, m_dwTempError);
    return false;
  }

  //Set the send timeout if necessary
  if ((m_dwSendTimeout != 0) && !InternetSetOption(m_hFTPConnection, INTERNET_OPTION_DATA_SEND_TIMEOUT, &m_dwSendTimeout, sizeof(m_dwSendTimeout)))
  {
    const DWORD dwError{GetLastError()};
    sl.Unlock();
    ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed in call to InternetSetOption to send send timeout, Error:%u\n"), dwError);
    HandleError(IDS_FTPTRANSFER_GENERIC_ERROR, dwError);
    return false;
  }
  //Set the receive timeout if necessary
  if ((m_dwReceiveTimeout != 0) && !InternetSetOption(m_hFTPConnection, INTERNET_OPTION_DATA_RECEIVE_TIMEOUT, &m_dwReceiveTimeout, sizeof(m_dwReceiveTimeout)))
  {
    const DWORD dwError{GetLastError()};
    sl.Unlock();
    ATLTRACE(_T("CFTPTransferer::CreateConnection, Failed in call to InternetSetOption to set receive timeout, Error:%u\n"), dwError);
    HandleError(IDS_FTPTRANSFER_GENERIC_ERROR, dwError);
    return false;
  }

  return true;
}

bool CFTPTransferer::CreateRequest()
{
  //Serialize access to the handles
  ATL::CCritSecLock sl{m_csHandles.m_sec, true};

  //Open the FTP file
#pragma warning(suppress: 26477)
  ATLASSERT(m_hFTPFile == nullptr);
#pragma warning(suppress: 26490)
  m_hFTPFile = FtpOpenFile(m_hFTPConnection, m_sRemoteFile.c_str(), m_bDownload ? GENERIC_READ : GENERIC_WRITE, m_bBinary ? FTP_TRANSFER_TYPE_BINARY | INTERNET_FLAG_RELOAD | INTERNET_FLAG_DONT_CACHE : 
                           FTP_TRANSFER_TYPE_ASCII | INTERNET_FLAG_RELOAD | INTERNET_FLAG_DONT_CACHE, reinterpret_cast<DWORD_PTR>(this));
  if (m_hFTPFile == nullptr)
  {
    const DWORD dwLastError{::GetLastError()};
    sl.Unlock();
    ATLTRACE(_T("CFTPTransferer::_Transfer, Failed in call to FtpOpenFile, Error:%u\n"), dwLastError);
    HandleError(IDS_FTPTRANSFER_FAIL_OPEN_FILE, dwLastError);
    return false;
  }
  sl.Unlock();

  //Update the status to say that we are now uploading / downloading the file
  if (m_bDownload)
    SetStatus(IDS_FTPTRANSFER_RETRIEVING_FILE);
  else
    SetStatus(IDS_FTPTRANSFER_UPLOADING_FILE);

  return true;
}

bool CFTPTransferer::OnTransferData(_In_reads_bytes_(dwSize) const BYTE* pbyData, _In_ DWORD dwSize)
{
  if (m_bDownload)
  {
    //Write the data to file
    const HRESULT hr{m_LocalFile.Write(pbyData, dwSize)};
    if (FAILED(hr))
    {
      ATLTRACE(_T("CFTPTransferer::OnTransferData, An exception occurred while writing to the download file, Error:%08X\n"), hr);
      HandleError(IDS_FTPTRANSFER_ERROR_WRITEFILE, hr);
      return false;
    }
  }
  else
  {
#pragma warning(suppress: 26477)
    ATLASSUME(m_hFTPFile != nullptr);
    DWORD dwBytesWritten{0};
    if (!::InternetWriteFile(m_hFTPFile, pbyData, dwSize, &dwBytesWritten))
    {
      const DWORD dwLastError{::GetLastError()};
      ATLTRACE(_T("CFTPTransferDlg::TransferThread, Failed in call to InternetWriteFile, Error:%u\n"), dwLastError);
      HandleError(IDS_FTPTRANSFER_ERROR_WRITEFILE, dwLastError);
      return false;
    }
  }

  //Continue the download if we are not aborting
  return !IsAbort();
}

_Return_type_success_(return != false) bool CFTPTransferer::GetRemoteFileLength(_Out_ ULONGLONG& nFileSize) noexcept
{
  //What will be the return value (assume the worst)
  bool bSuccess{false};

  WIN32_FIND_DATA wfd{};
#pragma warning(suppress: 26490)
  HINTERNET hFind{::FtpFindFirstFile(m_hFTPConnection, m_sRemoteFile.c_str(), &wfd, INTERNET_FLAG_RELOAD | INTERNET_FLAG_DONT_CACHE, reinterpret_cast<DWORD_PTR>(this))};
  if (hFind)
  {
    bSuccess = true;

    //Set the file size output parameter
    nFileSize = (static_cast<ULONGLONG>(wfd.nFileSizeHigh) << 32) + wfd.nFileSizeLow;

    //Destroy the enumeration handle now that we are finished with it
    InternetCloseHandle(hFind);
  }

  return bSuccess;
}

bool CFTPTransferer::_Transfer()
{
  //Make sure our wininet handles are empty
  ReleaseHandles();

  //Try and open the file we will be downloading into / uploading from
  if (m_bDownload)
  {
    HRESULT hr{m_LocalFile.Create(m_sLocalFile.c_str(), GENERIC_WRITE, 0, OPEN_ALWAYS)};
    if (FAILED(hr))
    {
      ATLTRACE(_T("CFTPTransferer::_Transfer, Failed to open the file to download into / upload from, Error:%08X\n"), hr);
      HandleError(IDS_FTPTRANSFER_FAIL_LOCALFILE_OPEN1, hr);
      return false;
    }

    //Seek to the correct start position
    hr = m_LocalFile.Seek(0, FILE_BEGIN);
    if (SUCCEEDED(hr))
      hr = m_LocalFile.SetSize(0);
    if (FAILED(hr))
    {
      ATLTRACE(_T("CFTPTransferer::_Transfer, An exception occurred setting the file size, Error:%08X\n"), hr);
      HandleError(IDS_FTPTRANSFER_FAIL_FILE_SEEK, hr);
      return false;
    }
  }
  else
  {
    const HRESULT hr{m_LocalFile.Create(m_sLocalFile.c_str(), GENERIC_READ, 0, OPEN_ALWAYS)};
    if (FAILED(hr))
    {
      ATLTRACE(_T("CFTPTransferer::_Transfer, Failed to open the file to upload, Error:%08X\n"), hr);
      HandleError(IDS_FTPTRANSFER_FAIL_LOCALFILE_OPEN2, hr);
      return false;
    }
  }

  //Call the virtual functions which create the session and connection
  if (!CreateSession())
    return false;
  if (!CreateConnection())
    return false;

  //Get the length of the file on the FTP server
  ULONGLONG nFileSize{0};
  bool bGotFileSize{false};
  if (m_bDownload)
  {
    //Update the status control to reflect that we are getting the file information
    SetStatus(IDS_FTPTRANSFER_GETTING_FILE_INFORMATION);

    bGotFileSize = GetRemoteFileLength(nFileSize);
  }
  else
  {
    bGotFileSize = true;
    const HRESULT hr{m_LocalFile.GetSize(nFileSize)};
    if (FAILED(hr))
      return false;
  }

  //Call the virtual function to open the FTP file
  if (!CreateRequest())
    return false;

  //Now do the actual read of the file
  const ULONGLONG dwStartTicks{GetTickCount64()};
  DWORD dwBytesRead{0};
  std::vector<BYTE> readBuffer{m_dwReadBufferSize, std::allocator<BYTE>{}};
  ULONGLONG nTotalBytesRead{0};
  do
  {
    if (m_bDownload)
    {
      //Read from the remote file
#pragma warning(suppress: 26486)
      if (!::InternetReadFile(m_hFTPFile, readBuffer.data(), m_dwReadBufferSize, &dwBytesRead))
      {
        const DWORD dwLastError{::GetLastError()};
        ATLTRACE(_T("CFTPTransferer::_Transfer, Failed in call to InternetReadFile, Error:%u\n"), dwLastError);
        HandleError(IDS_FTPTRANSFER_ERROR_READFILE, dwLastError);
        return false;
      }
    }
    else
    {
      //Read the data from the local file
#pragma warning(suppress: 26486)
      const HRESULT hr{m_LocalFile.Read(readBuffer.data(), m_dwReadBufferSize, dwBytesRead)};
      if (FAILED(hr))
      {
        ATLTRACE(_T("CFTPTransferer::_Transfer, An exception occurred while reading the local file, Error:%08X\n"), hr);
        HandleError(IDS_FTPTRANSFER_ERROR_READFILE, hr);
        return false;
      }
    }

    if (dwBytesRead)
    {
      //Call the virtual function which handles the transfer
#pragma warning(suppress: 26486)
      if (!OnTransferData(readBuffer.data(), dwBytesRead))
        return false;

      //Increment the total number of bytes read
      nTotalBytesRead += dwBytesRead;

      //For bandwidth throttling
      if (m_dbLimit > 0)
      {
        const auto t{static_cast<double>(GetTickCount64() - dwStartTicks)};
        const auto q{static_cast<double>(nTotalBytesRead) / t};

        if (q > m_dbLimit)
          Sleep(static_cast<int>(((q*t)/m_dbLimit) - t));
      }

      //Call the virtual function which reports progress as the download is occurring
      OnProgress(nTotalBytesRead, bGotFileSize, nFileSize);
    }
  }
  while (dwBytesRead && !IsAbort());

  //We're finished
  return true;
}

#pragma warning(suppress: 26440)
void CFTPTransferer::OnProgress(_In_ ULONGLONG nTotalBytesRead, _In_ bool bGotMaxLength, _In_ ULONGLONG nTotalBytesToRead)
{
#ifndef _DEBUG
  UNREFERENCED_PARAMETER(nTotalBytesRead);
  UNREFERENCED_PARAMETER(nTotalBytesToRead);
#endif

  //Just trace out the progress, derived classes are free to implement something more useful
  if (bGotMaxLength)
    ATLTRACE(_T("Transferred %I64u of %I64u bytes\n"), nTotalBytesRead, nTotalBytesToRead);
  else
    ATLTRACE(_T("Transferred %I64u bytes\n"), nTotalBytesRead);
}

bool CFTPTransferer::IsAbort() noexcept
{
  return m_bAbort;
}

void CFTPTransferer::SetAbort()
{
  m_bAbort = true;

  //Also release the Wininet handles, this should cause any blocking calls to Wininet to fail with errors
  ReleaseHandles();
}

void CFTPTransferer::HandleError(_In_ UINT nID, _In_ DWORD dwError)
{
  ATL::CAtlString sErrorMsg;
  if (!sErrorMsg.LoadString(nID))
    return;

  //Just trace out the error, derived classes are free to implement something more useful
  ATL::CAtlString sError;
  if (dwError == ERROR_INTERNET_EXTENDED_ERROR)
  {
    DWORD dwInetError{0};
    DWORD dwSize{0};
    ::InternetGetLastResponseInfo(&dwInetError, nullptr, &dwSize);
    TCHAR* pszResponse{sError.GetBuffer(dwSize)};
    ::InternetGetLastResponseInfo(&dwInetError, pszResponse, &dwSize);
    sError.ReleaseBuffer();
  }
  else
    sError = GetErrorMessage(dwError).c_str();
  sErrorMsg.Replace(_T("%1"), sError);
  ATLTRACE(_T("%s\n"), sErrorMsg.GetString());
}

void CFTPTransferer::SetStatus(_In_ UINT nID)
{
  //Just load up the string and pass to the other version of the function
  ATL::CAtlString sStatus;
  if (sStatus.LoadString(nID))
    SetStatus(sStatus);
}

void CFTPTransferer::SetStatus(_In_ UINT nID, _In_z_ LPCTSTR lpsz1)
{
  //Just form out the string and pass to the other version of the function
  ATL::CAtlString sStatus;
  if (!sStatus.LoadString(nID))
    return;
  sStatus.Replace(_T("%1"), lpsz1);
  SetStatus(sStatus);
}

#pragma warning(suppress: 26440)
void CFTPTransferer::SetStatus(_In_z_ LPCTSTR pszStatus)
{
#ifndef _DEBUG
  UNREFERENCED_PARAMETER(pszStatus);
#endif

  //Just trace out the string, derived classes are free to implement something more useful
  ATLTRACE(_T("CFTPTransferer::SetStatus, %s\n"), pszStatus);
}

CFTPTransferer::String CFTPTransferer::GetErrorMessage(_In_ DWORD dwError)
{
  //use the Wininet dll also to find error strings
#pragma warning(suppress: 26462)
  HMODULE hWininet{GetModuleHandle(_T("WININET.DLL"))};

  ATL::CAtlString rVal;
#pragma warning(suppress: 26429)
  LPTSTR pstrError{rVal.GetBuffer(4096)};
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS |
                                     FORMAT_MESSAGE_FROM_HMODULE, hWininet, dwError, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), 
                                     reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
  if (dwReturn == 0)
    *pstrError = _T('\0');
  else
  {
    ATL::Checked::tcsncpy_s(pstrError, 4096, lpBuffer, _TRUNCATE);
    LocalFree(lpBuffer);
  }
  rVal.ReleaseBuffer();

  return String{rVal};
}

void CALLBACK CFTPTransferer::_OnWininetStatusCallBack(_In_ HINTERNET hInternet, _In_ DWORD_PTR dwContext, _In_ DWORD dwInternetStatus,
                                                       _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength)
{
  //Convert from the SDK C world to the C++ world
#pragma warning(suppress: 26490)
  auto pThis{reinterpret_cast<CFTPTransferer*>(dwContext)};
#pragma warning(suppress: 26477)
  ATLASSUME(pThis != nullptr);
#pragma warning(suppress: 26486)
  pThis->OnWininetStatusCallBack(hInternet, dwInternetStatus, lpvStatusInformation, dwStatusInformationLength);
}

#pragma warning(suppress: 26440)
void CFTPTransferer::OnWininetStatusCallBack(_In_ HINTERNET /*hInternet*/, _In_ DWORD /*dwInternetStatus*/, _In_opt_ LPVOID /*lpvStatusInformation*/, _In_ DWORD /*dwStatusInformationLength*/)
{
  //The base class does nothing with this information, derived classes are free to implement something more useful
}

unsigned CFTPTransferer::_MakeConnectionThread(_In_ void* pParam) noexcept
{
  //Convert from the SDK world to the C++ world
  auto pThis{static_cast<CFTPTransferer*>(pParam)};
#pragma warning(suppress: 26477 26496)
  ATLASSUME(pThis != nullptr);
  return pThis->MakeConnection();
}

unsigned CFTPTransferer::MakeConnection() noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(m_hFTPConnection == nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(m_sServer.length());

  if (m_sUserName.length())
#pragma warning(suppress: 26490)
    m_hFTPConnection = ::InternetConnect(m_hInternetSession, m_sServer.c_str(), m_nPort, m_sUserName.c_str(), 
                                         m_sPassword.c_str(), INTERNET_SERVICE_FTP, m_bPasv ? INTERNET_FLAG_PASSIVE : 0, reinterpret_cast<DWORD_PTR>(this));
  else
#pragma warning(suppress: 26490)
    m_hFTPConnection = ::InternetConnect(m_hInternetSession, m_sServer.c_str(), m_nPort, nullptr,
                                         nullptr, INTERNET_SERVICE_FTP, m_bPasv ? INTERNET_FLAG_PASSIVE : 0, reinterpret_cast<DWORD_PTR>(this));
  if (m_hFTPConnection == nullptr)
  {
    //Hive away this threads error code as we have use for it in the main thread
    m_dwTempError = ::GetLastError();
  }

  return m_hFTPConnection ? 0 : 1;
}
