/*
Module : FTPTransferer.h
Purpose: Defines the interface for an C++ class which provides a wrapping for FTP transfers in a 
         synchronous manner via Wininet

Copyright (c) 2006 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __FTPTRANSFERER_H__
#define __FTPTRANSFERER_H__

#ifndef CFTPTRANSFERER_EXT_CLASS
#define CFTPTRANSFERER_EXT_CLASS
#endif //#ifndef CFTPTRANSFERER_EXT_CLASS


//////////////////// Includes /////////////////////////////////////////////////

#ifndef __ATLBASE_H__
#pragma message("To avoid this message, please put atlbase.h in your pre compiled header (normally stdafx.h)")
#include <atlbase.h>
#endif //#ifndef __ATLBASE_H__

#ifndef _ATOMIC_
#pragma message("To avoid this message, please put atomic in your pre compiled header (usually stdafx.h)")
#include <atomic>
#endif //#ifndef _ATOMIC_

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (usually stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef __ATLFILE_H__
#pragma message("To avoid this message, please atlfile.h in your pre compiled header (usually stdafx.h)")
#include <atlfile.h>
#endif //#ifndef __ATLFILE_H__

#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please atlstr.h in your pre compiled header (usually stdafx.h)")
#include <atlstr.h>
#endif //#ifndef __ATLSTR_H__


//////////////////// Classes //////////////////////////////////////////////////

//The synchronous downloading API which CFTPTransferDlg incorporates. If you do not want
//any UI then you can use this class directly
class CFTPTRANSFERER_EXT_CLASS CFTPTransferer
{
public:
//Constructors / Destructors
  CFTPTransferer() noexcept;
  CFTPTransferer(const CFTPTransferer&) = delete;
  CFTPTransferer(CFTPTransferer&&) = delete;
  virtual ~CFTPTransferer();

//Methods
  CFTPTransferer& operator=(const CFTPTransferer&) = delete;
  CFTPTransferer& operator=(CFTPTransferer&&) = delete;
  bool Transfer();
  bool DeleteDownloadedFile();
  void AttachSession(_In_ HINTERNET hInternetSession) noexcept;
  HINTERNET DetachSession() noexcept;

//Enums
  enum class ConnectionType
  {
    UsePreConfig,
    DirectToInternet,
    UseProxy
  };

//Typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE

//Member variables
  bool m_bDownload;                //true if it's a download, false if an upload
  String m_sServer;                //e.g. "ftp.microsoft.com"
  String m_sRemoteFile;            //e.g. "/pub/somefile.ext"
  String m_sLocalFile;             //e.g. "c:\temp\somefile.txt"
  String m_sUserName;              //Username to login to the server with
  String m_sPassword;              //password to login to the server with
  INTERNET_PORT m_nPort;           //If you want to change the port to make access on, by default it will be 21
  bool m_bBinary;                  //true if binary transfer, false for ascii transfer
  bool m_bPasv;                    //should a PASV connection be used
  double m_dbLimit;                //For bandwidth throttling, The value in Bytes / Second to limit the connection to
  DWORD m_dwConnectionTimeout;     //The connection timeout to use (in milliseconds), (Use a value of 0 to not specify a connection timeout)
  String m_sUserAgent;             //The User agent to use
  String m_sProxyServer;           //The proxy server to use
  ConnectionType m_ConnectionType; //The connection type to use
  DWORD m_dwReadBufferSize;        //The size of the receive buffer when we read the HTTP response
  DWORD m_dwSendTimeout;           //The send timeout to use (in milliseconds), (Use a value of 0 to not specify a send timeout)
  DWORD m_dwReceiveTimeout;        //The receive timeout to use (in milliseconds), (Use a value of 0 to not specify a receive timeout)

protected:
//Member variables
  HINTERNET m_hInternetSession;
  HINTERNET m_hFTPConnection;
  HINTERNET m_hFTPFile;
  std::atomic<bool> m_bAbort;
  ATL::CAtlFile m_LocalFile;
  ATL::CComAutoCriticalSection m_csHandles; //Critical section we use to serialize access to the Wininet handles
  DWORD m_dwTempError;

//Static methods
  static void CALLBACK _OnWininetStatusCallBack(_In_ HINTERNET hInternet, _In_ DWORD_PTR dwContext, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength);
  static String GetErrorMessage(_In_ DWORD dwError);
  static unsigned __stdcall _MakeConnectionThread(_In_ void* pParam) noexcept;

//Standard methods
  void ReleaseHandles();
  void SetAbort();
  bool _Transfer();

//Virtual methods
  virtual bool CreateSession();
  virtual bool CreateConnection();
  virtual bool CreateRequest();
  virtual unsigned MakeConnection() noexcept;
  virtual void OnWininetStatusCallBack(_In_ HINTERNET hInternet, _In_ DWORD dwInternetStatus, _In_opt_ LPVOID lpvStatusInformation, _In_ DWORD dwStatusInformationLength);
  virtual void HandleError(_In_ UINT nID, _In_ DWORD dwError);
  virtual void SetStatus(_In_ UINT nID);
  virtual void SetStatus(_In_ UINT nID, _In_z_ LPCTSTR lpsz1);
  virtual void SetStatus(_In_z_ LPCTSTR pszStatus);
  virtual bool IsAbort() noexcept;
  virtual void OnProgress(_In_ ULONGLONG nBytesRead, _In_ bool bGotFileSize, _In_ ULONGLONG nTotalBytesToRead);
  virtual bool OnTransferData(_In_reads_bytes_(dwSize) const BYTE* pbyData, _In_ DWORD dwSize);
  virtual _Return_type_success_(return != false) bool GetRemoteFileLength(_Out_ ULONGLONG& dwFileSize) noexcept;
};

#endif //#ifndef __FTPTRANSFERER_H__
